@@ -29,6 +29,7 @@
 #include "context.h"
 #include "ec-context.h"
 #include "ec-internal.h"
+#include "cipher.h" /* for GCRYECC_FLAG_LEAST_LEAK */
 
 extern void reverse_buffer (unsigned char *buffer, unsigned int length);
 
@@ -183,6 +184,30 @@ point_swap_cond (mpi_point_t d, mpi_point_t s, unsigned long swap,
 }
 
 
+/*
+ * Move the point value of A from B, when DIR is 1.
+ * Move the point value of A to B, when DIR is 0.
+ *
+ * For the use case when DIR is 0, it's actually dummy operations (the
+ * value copied into B is not used after the call).  The intention
+ * here is to be constant-time and to reduce possible EM signal/noise
+ * ratio (by decreasing signal and increasing noise).
+ *
+ * The word "tfr" comes from the mnemonic of Motorola 6809
+ * instruction, which does "transfer" a register value to another
+ * register.  "TFR A,B" means "Transfer A to B".
+ */
+static void
+point_tfr (mpi_point_t a, mpi_point_t b, unsigned long dir,
+           mpi_ec_t ctx)
+{
+  mpi_tfr (a->x, b->x, dir);
+  if (ctx->model != MPI_EC_MONTGOMERY)
+    mpi_tfr (a->y, b->y, dir);
+  mpi_tfr (a->z, b->z, dir);
+}
+
+
 /* Set the projective coordinates from POINT into X, Y, and Z.  If a
    coordinate is not required, X, Y, or Z may be passed as NULL.  */
 void
@@ -279,6 +304,9 @@ ec_mod (gcry_mpi_t w, mpi_ec_t ec)
     _gcry_mpi_mod_barrett (w, w, ec->t.p_barrett);
   else
     _gcry_mpi_mod (w, w, ec->p);
+
+  if ((ec->flags & GCRYECC_FLAG_LEAST_LEAK))
+    w->nlimbs = ec->p->nlimbs;
 }
 
 static void
@@ -304,12 +332,66 @@ ec_mulm (gcry_mpi_t w, gcry_mpi_t u, gcry_mpi_t v, mpi_ec_t ctx)
   ctx->mod (w, ctx);
 }
 
+static void
+ec_addm_lli (gcry_mpi_t w, gcry_mpi_t u, gcry_mpi_t v, mpi_ec_t ctx)
+{
+  mpi_limb_t cy1, cy2;
+
+  w->nlimbs = ctx->p->nlimbs;
+  cy1 = _gcry_mpih_add_lli (w->d, u->d, v->d, ctx->p->nlimbs);
+  cy2 = _gcry_mpih_sub_n (w->d, w->d, ctx->p->d, ctx->p->nlimbs);
+  _gcry_mpih_add_n_cond (w->d, w->d, ctx->p->d, ctx->p->nlimbs,
+                         ~cy1 &
+                         ((-cy2) >> (BITS_PER_MPI_LIMB - 1)));
+}
+
+static void
+ec_subm_lli (gcry_mpi_t w, gcry_mpi_t u, gcry_mpi_t v, mpi_ec_t ec)
+{
+  mpi_limb_t cy;
+
+  w->nlimbs = ec->p->nlimbs;
+  cy = _gcry_mpih_sub_n (w->d, u->d, v->d, ec->p->nlimbs);
+  _gcry_mpih_add_n_cond (w->d, w->d, ec->p->d, ec->p->nlimbs,
+                         (-cy) >> (BITS_PER_MPI_LIMB - 1));
+}
+
+static void
+ec_mulm_lli (gcry_mpi_t w, gcry_mpi_t u, gcry_mpi_t v, mpi_ec_t ctx)
+{
+  mpi_ptr_t wp;
+  mpi_ptr_t tmp_limb = NULL;
+  unsigned int tmp_limb_nlimbs = 0;
+  mpi_limb_t cy;
+  mpi_size_t usize, vsize, wsize;
+
+  usize = u->nlimbs;
+  vsize = v->nlimbs;
+  wsize = usize + vsize;
+  mpi_resize (w, wsize);
+  w->nlimbs = wsize;
+  if (w->d == u->d || w->d == v->d)
+    {
+      tmp_limb_nlimbs = wsize;
+      wp = tmp_limb = mpi_alloc_limb_space (tmp_limb_nlimbs, 0);
+    }
+  else
+    wp = w->d;
+  cy = _gcry_mpih_mul_lli (wp, u->d, usize, v->d, vsize);
+  wp[w->nlimbs - 1] = cy;
+  if (tmp_limb)
+    {
+      MPN_COPY (w->d, tmp_limb, wsize);
+      _gcry_mpi_free_limb_space (tmp_limb, tmp_limb_nlimbs);
+    }
+  ctx->mod (w, ctx);
+}
+
 /* W = 2 * U mod P.  */
 static void
 ec_mul2 (gcry_mpi_t w, gcry_mpi_t u, mpi_ec_t ctx)
 {
-  mpi_lshift (w, u, 1);
-  ctx->mod (w, ctx);
+  ec_addm (w, u, u, ctx);
 }
 
 static void
@@ -333,16 +415,6 @@ ec_pow2 (gcry_mpi_t w, const gcry_mpi_t b, mpi_ec_t ctx)
 }
 
 
-/* Shortcut for
-     ec_powm (B, B, mpi_const (MPI_C_THREE), ctx);
-   for easier optimization.  */
-static void
-ec_pow3 (gcry_mpi_t w, const gcry_mpi_t b, mpi_ec_t ctx)
-{
-  mpi_powm (w, b, mpi_const (MPI_C_THREE), ctx->p);
-}
-
-
 static void
 ec_invm (gcry_mpi_t x, gcry_mpi_t a, mpi_ec_t ctx)
 {
@@ -581,16 +653,12 @@ ec_pow2_448 (gcry_mpi_t w, const gcry_mpi_t b, mpi_ec_t ctx)
 static void
 ec_secp256k1_mod (gcry_mpi_t w, mpi_ec_t ctx)
 {
-  mpi_size_t wsize = (256 + BITS_PER_MPI_LIMB - 1) / BITS_PER_MPI_LIMB;
-  mpi_limb_t n[wsize + 1];
-  mpi_limb_t s[wsize + 1];
+  mpi_limb_t s[(256 + BITS_PER_MPI_LIMB - 1) / BITS_PER_MPI_LIMB + 1];
+  mpi_limb_t n[DIM(s)];
+  const mpi_size_t wsize = DIM(s) - 1;
   mpi_limb_t cy, borrow;
   mpi_ptr_t wp;
 
-  MPN_NORMALIZE (w->d, w->nlimbs);
-  if (w->nlimbs > 2 * 256 / BITS_PER_MPI_LIMB)
-    log_bug ("W must be less than m^2\n");
-
   RESIZE_AND_CLEAR_IF_NEEDED (w, wsize * 2);
 
   wp = w->d;
@@ -626,7 +694,8 @@ ec_secp256k1_mod (gcry_mpi_t w, mpi_ec_t ctx)
 		 mpih_limb_is_not_zero (cy) | mpih_limb_is_zero (borrow));
 
   w->nlimbs = wsize;
-  MPN_NORMALIZE (wp, w->nlimbs);
+  if (!(ctx->flags & GCRYECC_FLAG_LEAST_LEAK))
+    MPN_NORMALIZE (wp, w->nlimbs);
 }
 
 
@@ -743,8 +812,8 @@ ec_get_a_is_pminus3 (mpi_ec_t ec)
     {
       ec->t.valid.a_is_pminus3 = 1;
       tmp = mpi_alloc_like (ec->p);
-      mpi_sub_ui (tmp, ec->p, 3);
-      ec->t.a_is_pminus3 = !mpi_cmp (ec->a, tmp);
+      mpi_sub (tmp, ec->p, ec->a);
+      ec->t.a_is_pminus3 = !mpi_cmp_ui (tmp, 3);
       mpi_free (tmp);
     }
 
@@ -857,7 +926,7 @@ ec_p_init (mpi_ec_t ctx, enum gcry_mpi_ec_models model,
           if (!match_p)
             continue;
 
-          for (j=0; i< DIM(ctx->t.scratch) && bad_points_table[i][j]; j++)
+          for (j=0; j < DIM(ctx->t.scratch) && bad_points_table[i][j]; j++)
             ctx->t.scratch[j] = scanval (bad_points_table[i][j]);
         }
     }
@@ -902,26 +971,28 @@ ec_p_init (mpi_ec_t ctx, enum gcry_mpi_ec_models model,
           ctx->mul2 = field_table[i].mul2 ? field_table[i].mul2 : ctx->mul2;
           ctx->pow2 = field_table[i].pow2 ? field_table[i].pow2 : ctx->pow2;
           ctx->mod = field_table[i].mod ? field_table[i].mod : ctx->mod;
-
-	  if (ctx->a)
-	    {
-	      mpi_resize (ctx->a, ctx->p->nlimbs);
-	      ctx->a->nlimbs = ctx->p->nlimbs;
-	    }
-
-	  if (ctx->b)
-	    {
-	      mpi_resize (ctx->b, ctx->p->nlimbs);
-	      ctx->b->nlimbs = ctx->p->nlimbs;
-	    }
-
-          for (i=0; i< DIM(ctx->t.scratch) && ctx->t.scratch[i]; i++)
-            ctx->t.scratch[i]->nlimbs = ctx->p->nlimbs;
-
           break;
         }
     }
 
+  if (field_table[i].p || (flags & GCRYECC_FLAG_LEAST_LEAK))
+    {
+      if (ctx->a)
+        {
+          mpi_resize (ctx->a, ctx->p->nlimbs);
+          ctx->a->nlimbs = ctx->p->nlimbs;
+        }
+
+      if (ctx->b)
+        {
+          mpi_resize (ctx->b, ctx->p->nlimbs);
+          ctx->b->nlimbs = ctx->p->nlimbs;
+        }
+
+      for (i=0; i< DIM(ctx->t.scratch) && ctx->t.scratch[i]; i++)
+        ctx->t.scratch[i]->nlimbs = ctx->p->nlimbs;
+    }
+
   /* Prepare for fast reduction.  */
   /* FIXME: need a test for NIST values.  However it does not gain us
      any real advantage, for 384 bits it is actually slower than using
@@ -1025,7 +1096,7 @@ _gcry_mpi_ec_p_new (gcry_ctx_t *r_ctx,
   if (!p || !a)
     return GPG_ERR_EINVAL;
 
-  ctx = _gcry_ctx_alloc (CONTEXT_TYPE_EC, sizeof *ec, ec_deinit);
+  ctx = _gcry_ctx_alloc (CONTEXT_TYPE_EC, sizeof *ec, ec_deinit, NULL);
   if (!ctx)
     return gpg_err_code_from_syserror ();
   ec = _gcry_ctx_get_pointer (ctx, CONTEXT_TYPE_EC);
@@ -1127,6 +1198,7 @@ _gcry_mpi_ec_get_affine (gcry_mpi_t x, gcry_mpi_t y, mpi_point_t point,
     case MPI_EC_WEIERSTRASS: /* Using Jacobian coordinates.  */
       {
         gcry_mpi_t z1, z2, z3;
+        int saved_flags;
 
 	if (!mpi_cmp_ui (point->z, 1))
 	  {
@@ -1137,22 +1209,34 @@ _gcry_mpi_ec_get_affine (gcry_mpi_t x, gcry_mpi_t y, mpi_point_t point,
 	    return 0;
 	  }
 
+        /* Since we use LLI computations, the LL flag should be set.  */
+        saved_flags = ctx->flags;
+        ctx->flags |= GCRYECC_FLAG_LEAST_LEAK;
+
         z1 = mpi_new (0);
         z2 = mpi_new (0);
-        ec_invm (z1, point->z, ctx);  /* z1 = z^(-1) mod p  */
-        ec_mulm (z2, z1, z1, ctx);    /* z2 = z^(-2) mod p  */
+        ec_invm (z1, point->z, ctx);   /* z1 = z^(-1) mod p  */
+        ec_mulm_lli (z2, z1, z1, ctx); /* z2 = z^(-2) mod p  */
 
         if (x)
-          ec_mulm (x, point->x, z2, ctx);
+          {
+            mpi_resize (point->x, ctx->p->nlimbs);
+            point->x->nlimbs = ctx->p->nlimbs;
+            ec_mulm_lli (x, point->x, z2, ctx);
+          }
 
         if (y)
           {
+            mpi_resize (point->y, ctx->p->nlimbs);
+            point->y->nlimbs = ctx->p->nlimbs;
             z3 = mpi_new (0);
-            ec_mulm (z3, z2, z1, ctx);      /* z3 = z^(-3) mod p  */
-            ec_mulm (y, point->y, z3, ctx);
+            ec_mulm_lli (z3, z2, z1, ctx); /* z3 = z^(-3) mod p  */
+            ec_mulm_lli (y, point->y, z3, ctx);
             mpi_free (z3);
           }
 
+        ctx->flags = saved_flags;
+
         mpi_free (z2);
         mpi_free (z1);
       }
@@ -1245,7 +1329,8 @@ dup_point_weierstrass (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
           /*                          T2: used for the right term.  */
           ec_pow2 (t1, point->z, ctx);
           ec_subm (l1, point->x, t1, ctx);
-          ec_mulm (l1, l1, mpi_const (MPI_C_THREE), ctx);
+          ec_addm (t3, l1, l1, ctx);
+          ec_addm (l1, l1, t3, ctx);
           ec_addm (t2, point->x, t1, ctx);
           ec_mulm (l1, l1, t2, ctx);
         }
@@ -1254,8 +1339,10 @@ dup_point_weierstrass (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
           /* L1 = 3X^2 + aZ^4 */
           /*                          T1: used for aZ^4. */
           ec_pow2 (l1, point->x, ctx);
-          ec_mulm (l1, l1, mpi_const (MPI_C_THREE), ctx);
-          ec_powm (t1, point->z, mpi_const (MPI_C_FOUR), ctx);
+          ec_addm (t3, l1, l1, ctx);
+          ec_addm (l1, l1, t3, ctx);
+          ec_pow2 (t1, point->z, ctx);
+          ec_pow2 (t1, t1, ctx);
           ec_mulm (t1, t1, ctx->a, ctx);
           ec_addm (l1, l1, t1, ctx);
         }
@@ -1267,7 +1354,8 @@ dup_point_weierstrass (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
       /*                              T2: used for Y2; required later. */
       ec_pow2 (t2, point->y, ctx);
       ec_mulm (l2, t2, point->x, ctx);
-      ec_mulm (l2, l2, mpi_const (MPI_C_FOUR), ctx);
+      ec_addm (l2, l2, l2, ctx);
+      ec_addm (l2, l2, l2, ctx);
 
       /* X3 = L1^2 - 2L2 */
       /*                              T1: used for L2^2. */
@@ -1278,7 +1366,9 @@ dup_point_weierstrass (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
       /* L3 = 8Y^4 */
       /*                              T2: taken from above. */
       ec_pow2 (t2, t2, ctx);
-      ec_mulm (l3, t2, mpi_const (MPI_C_EIGHT), ctx);
+      ec_addm (l3, t2, t2, ctx);
+      ec_addm (l3, l3, l3, ctx);
+      ec_addm (l3, l3, l3, ctx);
 
       /* Y3 = L1(L2 - X3) - L3 */
       ec_subm (y3, l2, x3, ctx);
@@ -1298,6 +1388,85 @@ dup_point_weierstrass (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
 }
 
 
+/*  RESULT = 2 * POINT  (Weierstrass version). */
+static void
+dup_point_weierstrass_lli (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
+{
+#define x3 (result->x)
+#define y3 (result->y)
+#define z3 (result->z)
+#define t1 (ctx->t.scratch[0])
+#define t2 (ctx->t.scratch[1])
+#define t3 (ctx->t.scratch[2])
+#define l1 (ctx->t.scratch[3])
+#define l2 (ctx->t.scratch[4])
+#define l3 (ctx->t.scratch[5])
+
+  if (ec_get_a_is_pminus3 (ctx))  /* Use the faster case.  */
+    {
+      /* L1 = 3(X - Z^2)(X + Z^2) */
+      /*                          T1: used for Z^2. */
+      /*                          T2: used for the right term.  */
+      ec_mulm_lli (t1, point->z, point->z, ctx);
+      ec_subm_lli (l1, point->x, t1, ctx);
+      ec_addm_lli (t3, l1, l1, ctx);
+      ec_addm_lli (l1, l1, t3, ctx);
+      ec_addm_lli (t2, point->x, t1, ctx);
+      ec_mulm_lli (l1, l1, t2, ctx);
+    }
+  else /* Standard case. */
+    {
+      /* L1 = 3X^2 + aZ^4 */
+      /*                          T1: used for aZ^4. */
+      ec_mulm_lli (l1, point->x, point->x, ctx);
+      ec_addm_lli (t3, l1, l1, ctx);
+      ec_addm_lli (l1, l1, t3, ctx);
+      ec_mulm_lli (t1, point->z, point->z, ctx);
+      ec_mulm_lli (t1, t1, t1, ctx);
+      ec_mulm_lli (t1, t1, ctx->a, ctx);
+      ec_addm_lli (l1, l1, t1, ctx);
+    }
+  /* Z3 = 2YZ */
+  ec_mulm_lli (z3, point->y, point->z, ctx);
+  ec_addm_lli (z3, z3, z3, ctx);
+
+  /* L2 = 4XY^2 */
+  /*                              T2: used for Y2; required later. */
+  ec_mulm_lli (t2, point->y, point->y, ctx);
+  ec_mulm_lli (l2, t2, point->x, ctx);
+  ec_addm_lli (l2, l2, l2, ctx);
+  ec_addm_lli (l2, l2, l2, ctx);
+
+  /* X3 = L1^2 - 2L2 */
+  /*                              T1: used for L2^2. */
+  ec_mulm_lli (x3, l1, l1, ctx);
+  ec_addm_lli (t1, l2, l2, ctx);
+  ec_subm_lli (x3, x3, t1, ctx);
+
+  /* L3 = 8Y^4 */
+  /*                              T2: taken from above. */
+  ec_mulm_lli (t2, t2, t2, ctx);
+  ec_addm_lli (l3, t2, t2, ctx);
+  ec_addm_lli (l3, l3, l3, ctx);
+  ec_addm_lli (l3, l3, l3, ctx);
+
+  /* Y3 = L1(L2 - X3) - L3 */
+  ec_subm_lli (y3, l2, x3, ctx);
+  ec_mulm_lli (y3, y3, l1, ctx);
+  ec_subm_lli (y3, y3, l3, ctx);
+
+#undef x3
+#undef y3
+#undef z3
+#undef t1
+#undef t2
+#undef t3
+#undef l1
+#undef l2
+#undef l3
+}
+
+
 /*  RESULT = 2 * POINT  (Montgomery version). */
 static void
 dup_point_montgomery (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
@@ -1402,6 +1571,83 @@ _gcry_mpi_ec_dup_point (mpi_point_t result, mpi_point_t point, mpi_ec_t ctx)
 }
 
 
+/* RESULT = P1 + P2  (Weierstrass version).
+ * P2 is represented by affine coordinate (X2,Y2).
+ * P1 is never same to P2.
+ * P1 may be at infinity, in this case, it results invalid RESULT.
+ * It's caller's responsibility not to use the invalid RESULT.  */
+static void
+add_points_weierstrass_a (mpi_point_t result,
+                          mpi_point_t p1,
+                          gcry_mpi_t x2, gcry_mpi_t y2,
+                          mpi_ec_t ctx)
+{
+#define x1 (p1->x    )
+#define y1 (p1->y    )
+#define z1 (p1->z    )
+#define x3 (result->x)
+#define y3 (result->y)
+#define z3 (result->z)
+#define l1 (ctx->t.scratch[0])
+#define l2 (ctx->t.scratch[1])
+#define l3 (ctx->t.scratch[2])
+#define l4 (ctx->t.scratch[3])
+#define l5 (ctx->t.scratch[4])
+#define l6 (ctx->t.scratch[5])
+#define l7 (ctx->t.scratch[6])
+#define l8 (ctx->t.scratch[7])
+#define l9 (ctx->t.scratch[8])
+
+  /* l2 = x2 z1^2  */
+  ec_mulm_lli (l5, z1, z1, ctx);
+  ec_mulm_lli (l2, l5, x2, ctx);
+  /* l3 = x1 - l2 */
+  ec_subm_lli (l3, x1, l2, ctx);
+  /* l5 = y2 z1^3  */
+  ec_mulm_lli (l5, z1, l5, ctx);
+  ec_mulm_lli (l5, l5, y2, ctx);
+  /* l6 = y1 - l5  */
+  ec_subm_lli (l6, y1, l5, ctx);
+
+  /* l7 = x1 + l2  */
+  ec_addm_lli (l7, x1, l2, ctx);
+  /* l8 = y1 + l5  */
+  ec_addm_lli (l8, y1, l5, ctx);
+  /* z3 = z1 l3  */
+  ec_mulm_lli (z3, z1, l3, ctx);
+  /* x3 = l6^2 - l7 l3^2  */
+  ec_mulm_lli (l1, l6, l6, ctx);
+  ec_mulm_lli (l2, l3, l3, ctx);
+  ec_mulm_lli (l4, l2, l7, ctx);
+  ec_subm_lli (x3, l1, l4, ctx);
+  /* l9 = l7 l3^2 - 2 x3  */
+  ec_addm_lli (l1, x3, x3, ctx);
+  ec_subm_lli (l9, l4, l1, ctx);
+  /* y3 = (l9 l6 - l8 l3^3)/2  */
+  ec_mulm_lli (l9, l9, l6, ctx);
+  ec_mulm_lli (l1, l3, l2, ctx);
+  ec_mulm_lli (l1, l1, l8, ctx);
+  ec_subm_lli (y3, l9, l1, ctx);
+  ec_mulm_lli (y3, y3, ec_get_two_inv_p (ctx), ctx);
+
+#undef x1
+#undef y1
+#undef z1
+#undef x3
+#undef y3
+#undef z3
+#undef l1
+#undef l2
+#undef l3
+#undef l4
+#undef l5
+#undef l6
+#undef l7
+#undef l8
+#undef l9
+}
+
+
 /* RESULT = P1 + P2  (Weierstrass version).*/
 static void
 add_points_weierstrass (mpi_point_t result,
@@ -1426,8 +1672,6 @@ add_points_weierstrass (mpi_point_t result,
 #define l7 (ctx->t.scratch[6])
 #define l8 (ctx->t.scratch[7])
 #define l9 (ctx->t.scratch[8])
-#define t1 (ctx->t.scratch[9])
-#define t2 (ctx->t.scratch[10])
 
   if ( (!mpi_cmp (x1, x2)) && (!mpi_cmp (y1, y2)) && (!mpi_cmp (z1, z2)) )
     {
@@ -1450,32 +1694,19 @@ add_points_weierstrass (mpi_point_t result,
     }
   else
     {
-      int z1_is_one = !mpi_cmp_ui (z1, 1);
-      int z2_is_one = !mpi_cmp_ui (z2, 1);
-
       /* l1 = x1 z2^2  */
       /* l2 = x2 z1^2  */
-      if (z2_is_one)
-        mpi_set (l1, x1);
-      else
-        {
-          ec_pow2 (l1, z2, ctx);
-          ec_mulm (l1, l1, x1, ctx);
-        }
-      if (z1_is_one)
-        mpi_set (l2, x2);
-      else
-        {
-          ec_pow2 (l2, z1, ctx);
-          ec_mulm (l2, l2, x2, ctx);
-        }
+      ec_pow2 (l4, z2, ctx);
+      ec_mulm (l1, l4, x1, ctx);
+      ec_pow2 (l5, z1, ctx);
+      ec_mulm (l2, l5, x2, ctx);
       /* l3 = l1 - l2 */
       ec_subm (l3, l1, l2, ctx);
       /* l4 = y1 z2^3  */
-      ec_powm (l4, z2, mpi_const (MPI_C_THREE), ctx);
+      ec_mulm (l4, z2, l4, ctx);
       ec_mulm (l4, l4, y1, ctx);
       /* l5 = y2 z1^3  */
-      ec_powm (l5, z1, mpi_const (MPI_C_THREE), ctx);
+      ec_mulm (l5, z1, l5, ctx);
       ec_mulm (l5, l5, y2, ctx);
       /* l6 = l4 - l5  */
       ec_subm (l6, l4, l5, ctx);
@@ -1505,18 +1736,18 @@ add_points_weierstrass (mpi_point_t result,
           ec_mulm (z3, z1, z2, ctx);
           ec_mulm (z3, z3, l3, ctx);
           /* x3 = l6^2 - l7 l3^2  */
-          ec_pow2 (t1, l6, ctx);
-          ec_pow2 (t2, l3, ctx);
-          ec_mulm (t2, t2, l7, ctx);
-          ec_subm (x3, t1, t2, ctx);
+          ec_pow2 (l1, l6, ctx);
+          ec_pow2 (l2, l3, ctx);
+          ec_mulm (l4, l2, l7, ctx);
+          ec_subm (x3, l1, l4, ctx);
           /* l9 = l7 l3^2 - 2 x3  */
-          ec_mul2 (t1, x3, ctx);
-          ec_subm (l9, t2, t1, ctx);
+          ec_mul2 (l1, x3, ctx);
+          ec_subm (l9, l4, l1, ctx);
           /* y3 = (l9 l6 - l8 l3^3)/2  */
           ec_mulm (l9, l9, l6, ctx);
-          ec_powm (t1, l3, mpi_const (MPI_C_THREE), ctx); /* fixme: Use saved value*/
-          ec_mulm (t1, t1, l8, ctx);
-          ec_subm (y3, l9, t1, ctx);
+          ec_mulm (l1, l3, l2, ctx);
+          ec_mulm (l1, l1, l8, ctx);
+          ec_subm (y3, l9, l1, ctx);
           ec_mulm (y3, y3, ec_get_two_inv_p (ctx), ctx);
         }
     }
@@ -1539,8 +1770,6 @@ add_points_weierstrass (mpi_point_t result,
 #undef l7
 #undef l8
 #undef l9
-#undef t1
-#undef t2
 }
 
 
@@ -1585,7 +1814,7 @@ add_points_edwards (mpi_point_t result,
 
   mpi_point_resize (result, ctx);
 
-  /* Compute: (X_3 : Y_3 : Z_3) = (X_1 : Y_1 : Z_1) + (X_2 : Y_2 : Z_3)  */
+  /* Compute: (X_3 : Y_3 : Z_3) = (X_1 : Y_1 : Z_1) + (X_2 : Y_2 : Z_2)  */
 
   /* A = Z1 · Z2 */
   ctx->mulm (A, Z1, Z2, ctx);
@@ -1655,6 +1884,99 @@ add_points_edwards (mpi_point_t result,
 }
 
 
+/* RESULT = P1 + P2  (Twisted Edwards version).
+ * P2 is represented by affine coordinate (X2,Y2).
+ * P1 is never same to P1.  */
+static void
+add_points_edwards_a (mpi_point_t result,
+                      mpi_point_t p1,
+                      gcry_mpi_t x2, gcry_mpi_t y2,
+                      mpi_ec_t ctx)
+{
+#define X1 (p1->x)
+#define Y1 (p1->y)
+#define Z1 (p1->z)
+#define X2 x2
+#define Y2 y2
+#define X3 (result->x)
+#define Y3 (result->y)
+#define Z3 (result->z)
+#define B (ctx->t.scratch[1])
+#define C (ctx->t.scratch[2])
+#define D (ctx->t.scratch[3])
+#define E (ctx->t.scratch[4])
+#define F (ctx->t.scratch[5])
+#define G (ctx->t.scratch[6])
+#define tmp (ctx->t.scratch[7])
+
+  mpi_point_resize (result, ctx);
+
+  /* Compute: (X_3 : Y_3 : Z_3) = (X_1 : Y_1 : Z_1) + (X_2 : Y_2 : 1)  */
+
+  /* B = Z1^2 */
+  ctx->pow2 (B, Z1, ctx);
+
+  /* C = X1 · X2 */
+  ctx->mulm (C, X1, X2, ctx);
+
+  /* D = Y1 · Y2 */
+  ctx->mulm (D, Y1, Y2, ctx);
+
+  /* E = d · C · D */
+  ctx->mulm (E, ctx->b, C, ctx);
+  ctx->mulm (E, E, D, ctx);
+
+  /* F = B - E */
+  ctx->subm (F, B, E, ctx);
+
+  /* G = B + E */
+  ctx->addm (G, B, E, ctx);
+
+  /* X_3 = Z1 · F · ((X_1 + Y_1) · (X_2 + Y_2) - C - D) */
+  ctx->addm (tmp, X1, Y1, ctx);
+  ctx->addm (X3, X2, Y2, ctx);
+  ctx->mulm (X3, X3, tmp, ctx);
+  ctx->subm (X3, X3, C, ctx);
+  ctx->subm (X3, X3, D, ctx);
+  ctx->mulm (X3, X3, F, ctx);
+  ctx->mulm (X3, X3, Z1, ctx);
+
+  /* Y_3 = Z1 · G · (D - aC) */
+  if (ctx->dialect == ECC_DIALECT_ED25519)
+    {
+      ctx->addm (Y3, D, C, ctx);
+    }
+  else
+    {
+      ctx->mulm (Y3, ctx->a, C, ctx);
+      ctx->subm (Y3, D, Y3, ctx);
+    }
+  ctx->mulm (Y3, Y3, G, ctx);
+  ctx->mulm (Y3, Y3, Z1, ctx);
+
+  /* Z_3 = F · G */
+  ctx->mulm (Z3, F, G, ctx);
+
+
+#undef X1
+#undef Y1
+#undef Z1
+#undef X2
+#undef Y2
+#undef X3
+#undef Y3
+#undef Z3
+#undef A
+#undef B
+#undef C
+#undef D
+#undef E
+#undef F
+#undef G
+#undef tmp
+}
+
+
 /* Compute a step of Montgomery Ladder (only use X and Z in the point).
    Inputs:  P1, P2, and x-coordinate of DIF = P1 - P1.
    Outputs: PRD = 2 * P1 and  SUM = P1 + P2. */
@@ -1770,6 +2092,209 @@ _gcry_mpi_ec_sub_points (mpi_point_t result,
 }
 
 
+/* Compute scalar point multiplication with Montgomery Ladder.
+   Note that we don't use Y-coordinate in the points at all.
+   RESULT->Y will be filled by zero.  */
+static void
+montgomery_mul_point (mpi_point_t result,
+                      gcry_mpi_t scalar, mpi_point_t point,
+                      mpi_ec_t ctx)
+{
+  unsigned int nbits;
+  int j;
+  gcry_mpi_t z1;
+  mpi_point_struct p1, p2;
+  mpi_point_struct p1_, p2_;
+  mpi_point_t q1, q2, prd, sum;
+  unsigned long sw;
+  mpi_size_t rsize;
+  int scalar_copied = 0;
+
+  nbits = mpi_get_nbits (scalar);
+  point_init (&p1);
+  point_init (&p2);
+  point_init (&p1_);
+  point_init (&p2_);
+  mpi_set_ui (p1.x, 1);
+  mpi_free (p2.x);
+  p2.x  = mpi_copy (point->x);
+  mpi_set_ui (p2.z, 1);
+
+  if (mpi_is_opaque (scalar))
+    {
+      const unsigned int pbits = ctx->nbits;
+      gcry_mpi_t a;
+      unsigned int n;
+      unsigned char *raw;
+
+      scalar_copied = 1;
+
+      raw = _gcry_mpi_get_opaque_copy (scalar, &n);
+      if ((n+7)/8 != (pbits+7)/8)
+        log_fatal ("scalar size (%d) != prime size (%d)\n",
+                   (n+7)/8, (pbits+7)/8);
+
+      reverse_buffer (raw, (n+7)/8);
+      if ((pbits % 8))
+        raw[0] &= (1 << (pbits % 8)) - 1;
+      raw[0] |= (1 << ((pbits + 7) % 8));
+      raw[(pbits+7)/8 - 1] &= (256 - ctx->h);
+      a = mpi_is_secure (scalar) ? mpi_snew (pbits): mpi_new (pbits);
+      _gcry_mpi_set_buffer (a, raw, (n+7)/8, 0);
+      xfree (raw);
+
+      scalar = a;
+    }
+
+  mpi_point_resize (&p1, ctx);
+  mpi_point_resize (&p2, ctx);
+  mpi_point_resize (&p1_, ctx);
+  mpi_point_resize (&p2_, ctx);
+
+  mpi_resize (point->x, ctx->p->nlimbs);
+  point->x->nlimbs = ctx->p->nlimbs;
+
+  q1 = &p1;
+  q2 = &p2;
+  prd = &p1_;
+  sum = &p2_;
+
+  for (j=nbits-1; j >= 0; j--)
+    {
+      mpi_point_t t;
+
+      sw = mpi_test_bit (scalar, j);
+      point_swap_cond (q1, q2, sw, ctx);
+      montgomery_ladder (prd, sum, q1, q2, point->x, ctx);
+      point_swap_cond (prd, sum, sw, ctx);
+      t = q1;  q1 = prd;  prd = t;
+      t = q2;  q2 = sum;  sum = t;
+    }
+
+  mpi_clear (result->y);
+  sw = (nbits & 1);
+  point_swap_cond (&p1, &p1_, sw, ctx);
+
+  rsize = p1.z->nlimbs;
+  MPN_NORMALIZE (p1.z->d, rsize);
+  if (rsize == 0)
+    {
+      mpi_set_ui (result->x, 1);
+      mpi_set_ui (result->z, 0);
+    }
+  else
+    {
+      z1 = mpi_new (0);
+      ec_invm (z1, p1.z, ctx);
+      ec_mulm (result->x, p1.x, z1, ctx);
+      mpi_set_ui (result->z, 1);
+      mpi_free (z1);
+    }
+
+  point_free (&p1);
+  point_free (&p2);
+  point_free (&p1_);
+  point_free (&p2_);
+  if (scalar_copied)
+    _gcry_mpi_release (scalar);
+}
+
+
+/* Compute scalar point multiplication, Least Leak Intended.  */
+static void
+mpi_ec_mul_point_lli (mpi_point_t result,
+                      gcry_mpi_t scalar, mpi_point_t point1,
+                      mpi_ec_t ctx)
+{
+  unsigned int nbits;
+  int j;
+  mpi_point_struct tmppnt;
+  mpi_point_struct point_;
+  mpi_point_t point = &point_;
+
+  mpi_point_resize (point1, ctx);
+
+  /* Convert POINT1 into affine coordinate, so that we can use
+     add_points_*_a routine with affine coordinate.  */
+  point_init (point);
+  if (_gcry_mpih_cmp_ui (point1->z->d, ctx->p->nlimbs, 1))
+    {
+      gcry_mpi_t x, y;
+
+      x = mpi_new (0);
+      y = mpi_new (0);
+
+      if (_gcry_mpi_ec_get_affine (x, y, point1, ctx))
+        {
+          mpi_free (x);
+          mpi_free (y);
+          point_free (point);
+          point_set (result, point1);
+          return;
+        }
+
+      mpi_set (point->x, x);
+      mpi_set (point->y, y);
+      mpi_set_ui (point->z, 1);
+      mpi_free (x);
+      mpi_free (y);
+    }
+  else
+    point_set (point, point1);
+
+  mpi_point_resize (point, ctx);
+
+  nbits = mpi_get_nbits (scalar);
+  if (nbits < ctx->nbits)
+    nbits = ctx->nbits;
+
+  if (ctx->model == MPI_EC_WEIERSTRASS)
+    {
+      mpi_set_ui (result->x, 1);
+      mpi_set_ui (result->y, 1);
+      mpi_set_ui (result->z, 0);
+    }
+  else
+    {
+      mpi_set_ui (result->x, 0);
+      mpi_set_ui (result->y, 1);
+      mpi_set_ui (result->z, 1);
+    }
+
+  point_init (&tmppnt);
+  mpi_point_resize (result, ctx);
+  mpi_point_resize (&tmppnt, ctx);
+  if (ctx->model == MPI_EC_WEIERSTRASS)
+    {
+      for (j=nbits-1; j >= 0; j--)
+        {
+          unsigned long is_z_zero;
+
+          dup_point_weierstrass_lli (result, result, ctx);
+          is_z_zero = _gcry_mpih_cmp_ui (result->z->d, ctx->p->nlimbs, 0) == 0;
+          add_points_weierstrass_a (&tmppnt, result, point->x, point->y, ctx);
+          /* When P1 is O (at infinity), computation of
+             add_points_weierstrass_a is invalid, and RESULT is P2.  */
+          mpih_set_cond (tmppnt.x->d, point->x->d, ctx->p->nlimbs, is_z_zero);
+          mpih_set_cond (tmppnt.y->d, point->y->d, ctx->p->nlimbs, is_z_zero);
+          mpih_set_cond (tmppnt.z->d, point->z->d, ctx->p->nlimbs, is_z_zero);
+          point_tfr (result, &tmppnt, mpi_test_bit (scalar, j), ctx);
+        }
+    }
+  else /* MPI_EC_EDWARDS */
+    {
+      for (j=nbits-1; j >= 0; j--)
+        {
+          dup_point_edwards (result, result, ctx);
+          add_points_edwards_a (&tmppnt, result, point->x, point->y, ctx);
+          point_tfr (result, &tmppnt, mpi_test_bit (scalar, j), ctx);
+        }
+    }
+  point_free (&tmppnt);
+  point_free (point);
+}
+
+
 /* Scalar point multiplication - the main function for ECC.  It takes
    an integer SCALAR and a POINT as well as the usual context CTX.
    RESULT will be set to the resulting point. */
@@ -1790,176 +2315,52 @@ _gcry_mpi_ec_mul_point (mpi_point_t result,
       return;
     }
 
-  if (ctx->model == MPI_EC_EDWARDS
-      || (ctx->model == MPI_EC_WEIERSTRASS
-          && mpi_is_secure (scalar)))
+  if (ctx->model == MPI_EC_MONTGOMERY)
     {
+      montgomery_mul_point (result, scalar, point, ctx);
+      return;
+    }
+  else if ((ctx->flags & GCRYECC_FLAG_LEAST_LEAK))
+    {
+      mpi_ec_mul_point_lli (result, scalar, point, ctx);
+      return;
+    }
+  else if (mpi_is_secure (scalar))
+    {
+      ctx->flags |= GCRYECC_FLAG_LEAST_LEAK;
+      mpi_ec_mul_point_lli (result, scalar, point, ctx);
+      ctx->flags &= ~GCRYECC_FLAG_LEAST_LEAK;
+      return;
+    }
+  else if (ctx->model == MPI_EC_EDWARDS)
+    {
+      int j;
+      unsigned int nbits = mpi_get_nbits (scalar);
+
+      mpi_set_ui (result->x, 0);
+      mpi_set_ui (result->y, 1);
+      mpi_set_ui (result->z, 1);
+
+      mpi_point_resize (result, ctx);
+      mpi_point_resize (point, ctx);
+
       /* Simple left to right binary method.  Algorithm 3.27 from
        * {author={Hankerson, Darrel and Menezes, Alfred J. and Vanstone, Scott},
        *  title = {Guide to Elliptic Curve Cryptography},
        *  year = {2003}, isbn = {038795273X},
        *  url = {http://www.cacr.math.uwaterloo.ca/ecc/},
        *  publisher = {Springer-Verlag New York, Inc.}} */
-      unsigned int nbits;
-      int j;
-
-      if (mpi_cmp (scalar, ctx->p) >= 0)
-        nbits = mpi_get_nbits (scalar);
-      else
-        nbits = mpi_get_nbits (ctx->p);
-
-      if (ctx->model == MPI_EC_WEIERSTRASS)
-        {
-          mpi_set_ui (result->x, 1);
-          mpi_set_ui (result->y, 1);
-          mpi_set_ui (result->z, 0);
-        }
-      else
-        {
-          mpi_set_ui (result->x, 0);
-          mpi_set_ui (result->y, 1);
-          mpi_set_ui (result->z, 1);
-          mpi_point_resize (point, ctx);
-        }
-
-      if (mpi_is_secure (scalar))
-        {
-          /* If SCALAR is in secure memory we assume that it is the
-             secret key we use constant time operation.  */
-          mpi_point_struct tmppnt;
-
-          point_init (&tmppnt);
-          mpi_point_resize (result, ctx);
-          mpi_point_resize (&tmppnt, ctx);
-          for (j=nbits-1; j >= 0; j--)
-            {
-              _gcry_mpi_ec_dup_point (result, result, ctx);
-              _gcry_mpi_ec_add_points (&tmppnt, result, point, ctx);
-              point_swap_cond (result, &tmppnt, mpi_test_bit (scalar, j), ctx);
-            }
-          point_free (&tmppnt);
-        }
-      else
-        {
-          if (ctx->model == MPI_EC_EDWARDS)
-            {
-              mpi_point_resize (result, ctx);
-              mpi_point_resize (point, ctx);
-            }
-
-          for (j=nbits-1; j >= 0; j--)
-            {
-              _gcry_mpi_ec_dup_point (result, result, ctx);
-              if (mpi_test_bit (scalar, j))
-                _gcry_mpi_ec_add_points (result, result, point, ctx);
-            }
-        }
-      return;
-    }
-  else if (ctx->model == MPI_EC_MONTGOMERY)
-    {
-      unsigned int nbits;
-      int j;
-      mpi_point_struct p1_, p2_;
-      mpi_point_t q1, q2, prd, sum;
-      unsigned long sw;
-      mpi_size_t rsize;
-      int scalar_copied = 0;
-
-      /* Compute scalar point multiplication with Montgomery Ladder.
-         Note that we don't use Y-coordinate in the points at all.
-         RESULT->Y will be filled by zero.  */
-
-      nbits = mpi_get_nbits (scalar);
-      point_init (&p1);
-      point_init (&p2);
-      point_init (&p1_);
-      point_init (&p2_);
-      mpi_set_ui (p1.x, 1);
-      mpi_free (p2.x);
-      p2.x  = mpi_copy (point->x);
-      mpi_set_ui (p2.z, 1);
-
-      if (mpi_is_opaque (scalar))
-        {
-          const unsigned int pbits = ctx->nbits;
-          gcry_mpi_t a;
-          unsigned int n;
-          unsigned char *raw;
-
-          scalar_copied = 1;
-
-          raw = _gcry_mpi_get_opaque_copy (scalar, &n);
-          if ((n+7)/8 != (pbits+7)/8)
-            log_fatal ("scalar size (%d) != prime size (%d)\n",
-                       (n+7)/8, (pbits+7)/8);
-
-          reverse_buffer (raw, (n+7)/8);
-          if ((pbits % 8))
-            raw[0] &= (1 << (pbits % 8)) - 1;
-          raw[0] |= (1 << ((pbits + 7) % 8));
-          raw[(pbits+7)/8 - 1] &= (256 - ctx->h);
-          a = mpi_is_secure (scalar) ? mpi_snew (pbits): mpi_new (pbits);
-          _gcry_mpi_set_buffer (a, raw, (n+7)/8, 0);
-          xfree (raw);
-
-          scalar = a;
-        }
-
-      mpi_point_resize (&p1, ctx);
-      mpi_point_resize (&p2, ctx);
-      mpi_point_resize (&p1_, ctx);
-      mpi_point_resize (&p2_, ctx);
-
-      mpi_resize (point->x, ctx->p->nlimbs);
-      point->x->nlimbs = ctx->p->nlimbs;
-
-      q1 = &p1;
-      q2 = &p2;
-      prd = &p1_;
-      sum = &p2_;
-
       for (j=nbits-1; j >= 0; j--)
         {
-          mpi_point_t t;
-
-          sw = mpi_test_bit (scalar, j);
-          point_swap_cond (q1, q2, sw, ctx);
-          montgomery_ladder (prd, sum, q1, q2, point->x, ctx);
-          point_swap_cond (prd, sum, sw, ctx);
-          t = q1;  q1 = prd;  prd = t;
-          t = q2;  q2 = sum;  sum = t;
+          _gcry_mpi_ec_dup_point (result, result, ctx);
+          if (mpi_test_bit (scalar, j))
+            _gcry_mpi_ec_add_points (result, result, point, ctx);
         }
-
-      mpi_clear (result->y);
-      sw = (nbits & 1);
-      point_swap_cond (&p1, &p1_, sw, ctx);
-
-      rsize = p1.z->nlimbs;
-      MPN_NORMALIZE (p1.z->d, rsize);
-      if (rsize == 0)
-        {
-          mpi_set_ui (result->x, 1);
-          mpi_set_ui (result->z, 0);
-        }
-      else
-        {
-          z1 = mpi_new (0);
-          ec_invm (z1, p1.z, ctx);
-          ec_mulm (result->x, p1.x, z1, ctx);
-          mpi_set_ui (result->z, 1);
-          mpi_free (z1);
-        }
-
-      point_free (&p1);
-      point_free (&p2);
-      point_free (&p1_);
-      point_free (&p2_);
-      if (scalar_copied)
-        _gcry_mpi_release (scalar);
       return;
     }
 
+  /* The case of Weierstrass curve.  */
+
   x1 = mpi_alloc_like (ctx->p);
   y1 = mpi_alloc_like (ctx->p);
   h  = mpi_alloc_like (ctx->p);
@@ -1992,9 +2393,9 @@ _gcry_mpi_ec_mul_point (mpi_point_t result,
       mpi_free (z2);
       mpi_free (z3);
     }
-  z1 = mpi_copy (mpi_const (MPI_C_ONE));
+  z1 = mpi_set_ui (NULL, 1);
 
-  mpi_mul (h, k, mpi_const (MPI_C_THREE)); /* h = 3k */
+  mpi_mul_ui (h, k, 3); /* h = 3k */
   loops = mpi_get_nbits (h);
   if (loops < 2)
     {
@@ -2081,7 +2482,8 @@ _gcry_mpi_ec_curve_point (gcry_mpi_point_t point, mpi_ec_t ctx)
         /* y^2 == x^3 + a·x + b */
         ec_pow2 (y, y, ctx);
 
-        ec_pow3 (xxx, x, ctx);
+        ec_pow2 (xxx, x, ctx);
+        ec_mulm (xxx, xxx, x, ctx);
         ec_mulm (w, ctx->a, x, ctx);
         ec_addm (w, w, ctx->b, ctx);
         ec_addm (w, w, xxx, ctx);
@@ -2103,20 +2505,19 @@ _gcry_mpi_ec_curve_point (gcry_mpi_point_t point, mpi_ec_t ctx)
         /* We check if right hand is quadratic residue or not by
            Euler's criterion.  */
         /* CTX->A has (a-2)/4 and CTX->B has b^-1 */
-        ec_mulm (w, ctx->a, mpi_const (MPI_C_FOUR), ctx);
-        ec_addm (w, w, mpi_const (MPI_C_TWO), ctx);
+        mpi_mul_ui (w, ctx->a, 4);
+        mpi_add_ui (w, w, 2);
         ec_mulm (w, w, x, ctx);
         ec_pow2 (xx, x, ctx);
         ec_addm (w, w, xx, ctx);
-        ec_addm (w, w, mpi_const (MPI_C_ONE), ctx);
         ec_mulm (w, w, x, ctx);
+        ec_addm (w, w, x, ctx);
         ec_mulm (w, w, ctx->b, ctx);
 #undef xx
         /* Compute Euler's criterion: w^(p-1)/2 */
-#define p_minus1 y
-        ec_subm (p_minus1, ctx->p, mpi_const (MPI_C_ONE), ctx);
-        mpi_rshift (p_minus1, p_minus1, 1);
-        ec_powm (w, w, p_minus1, ctx);
+#define p_minus1_half y
+        mpi_rshift (p_minus1_half, ctx->p, 1); /* p is odd */
+        ec_powm (w, w, p_minus1_half, ctx);
 
         res = !mpi_cmp_ui (w, 1);
 #undef p_minus1
